import sys
from typing import Any, Callable, NewType, Type, TypeVar, cast
if sys.version_info.minor >= 7:
    from typing import Protocol
else:
    from typing_extensions import Protocol


# ===================================================
# StatefulClass(Protocol)
# ===================================================


class StatefulClass(Protocol):
    def __getstate__(self) -> dict:
        ...

    def __setstate__(self, state: dict) -> None:
        ...


# ===================================================
# BoundMethodType(Protocol)
# ===================================================


BoundMethodInstanceType = NewType("BoundMethodInstanceType", object)
BoundMethodClassType = Type[BoundMethodInstanceType]
BoundMethodReturnType = TypeVar("BoundMethodReturnType")


class BoundMethodType(Protocol):
    __name__: str
    __self__: BoundMethodInstanceType

    def __call__(self, *args: Any, **kwargs: Any) -> BoundMethodReturnType:
        ...


def type_bound_method(bound_method: Callable[..., BoundMethodReturnType]) -> BoundMethodType:
    return cast(BoundMethodType, bound_method)


# ===================================================
# --- END ---
# ===================================================

if __name__ == "__main__":

    def return_instance_of_bound_method(method: BoundMethodType) -> BoundMethodInstanceType:
        return method.__self__

    def return_class_of_bound_method(method: BoundMethodType) -> BoundMethodClassType:
        return method.__self__.__class__

    def run_bounded_method(method: BoundMethodType, *args: Any, **kwargs: Any) -> BoundMethodReturnType:
        return method(*args, **kwargs)

    class MyClass:
        def __init__(self) -> None:
            pass

        def my_method(self) -> bool:
            return True

    # doesnt work out of the box...
    my_bound_method = MyClass().my_method
    print(return_instance_of_bound_method(my_bound_method))
    print(return_class_of_bound_method(my_bound_method))
    print(run_bounded_method(my_bound_method))

    # you need to use typing.cast() for it to work...
    my_bound_method = type_bound_method(MyClass().my_method)
    print(return_instance_of_bound_method(my_bound_method))
    print(return_class_of_bound_method(my_bound_method))
    print(run_bounded_method(my_bound_method))

    # let's try one more example
    from pathlib import Path
    my_bound_method = type_bound_method(Path().joinpath)
    print(return_instance_of_bound_method(my_bound_method))
    print(return_class_of_bound_method(my_bound_method))
    print(run_bounded_method(my_bound_method, "some_arg", "and_another"))



import logging

from pandas.io.sql import SQLDatabase, SQLiteDatabase

from maurice.patchers.core import patch_method_with_caching

logger = logging.getLogger(__name__)


def caching_patch_pandas_db() -> None:
    logger.debug("Patching pandas SQLDatabase and SQLiteDatabase (CACHING)...")
    dbs = (SQLDatabase, SQLiteDatabase)
    for cls in dbs:
        patch_method_with_caching(name="read_query", cls=cls, save_state=False)



from maurice.patchers.pandas import caching_patch_pandas_db
from maurice.patchers.sklearn import caching_patch_sklearn_estimators


def patch() -> None:
    caching_patch_sklearn_estimators()
    caching_patch_pandas_db()



import logging

from sklearn.utils import all_estimators

from maurice.patchers.core import patch_method_with_caching

logger = logging.getLogger(__name__)


def caching_patch_sklearn_estimators() -> None:
    logger.debug("Patching sklearn estimators (CACHING)...")
    for cls in tuple(zip(*all_estimators()))[1]:
        patch_method_with_caching(name="fit", cls=cls, save_state=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['maurice', 'maurice.patchers']

package_data = \
{'': ['*']}

install_requires = \
['dill>=0.3.1,<0.4.0', 'wrapt>=1.12.1,<2.0.0']

extras_require = \
{'all': ['scikit-learn', 'tensorflow>=2.4.0,<3.0.0', 'pandas>=1.0.0,<2.0.0'],
 'pandas': ['pandas>=1.0.0,<2.0.0'],
 'sklearn': ['scikit-learn'],
 'tensorflow': ['tensorflow>=2.4.0,<3.0.0']}

setup_kwargs = {
    'name': 'maurice',
    'version': '0.1.12',
    'description': 'Ship better machine learning projects, faster!',
    'long_description': '# Maurice\n\n\n## Installation\n\n### Using pypi\n\n```shell\npip install -U maurice\n```\n\n## Integrate it with your workflow\n\n### Automatic caching\n\nAdd any of the following patches at the top of your scripts\n\nCache SQL queries executed from the pandas library\n```python\nfrom maurice.patchers import caching_patch_pandas_db; caching_patch_pandas_db()\nimport pandas as pd\n\n\ndf = pd.read_sql_query(con=your_connection, sql="select * from your_table")\n```\n\nCache `.fit()` calls from any `sklearn` Estimator (...)\n```python\nfrom maurice.patchers import caching_patch_sklearn_estimators; caching_patch_sklearn_estimators()\n\n...\n```',
    'author': 'Tomas Pereira de Vasconcelos',
    'author_email': 'tomasvasconcelos1@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tpvasconcelos/maurice',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import logging

from tensorflow.python.keras.models import Model

from maurice.patchers.core import patch_method_with_caching

logger = logging.getLogger(__name__)


def caching_patch_keras_models() -> None:
    logger.debug("Patching Keras' Model (CACHING)...")
    patch_method_with_caching(name="fit", cls=Model, save_state=True)



import logging
from abc import ABCMeta, abstractmethod
from functools import partial

import dill
import wrapt

from maurice.caching import CACHE_DIR
from maurice.types import BoundMethodClassType, BoundMethodInstanceType, BoundMethodReturnType, BoundMethodType
from maurice.utils import hash_any

logger = logging.getLogger(__name__)


class BaseMethodWrapper(metaclass=ABCMeta):
    def __init__(self, method: BoundMethodType, args: tuple, kwargs: dict):
        self._method: BoundMethodType = method
        self._args: tuple = args
        self._kwargs: dict = kwargs

    @abstractmethod
    def run_wrapped(self) -> BoundMethodReturnType:
        pass


class CachingMethodWrapper(BaseMethodWrapper):
    def __init__(self, method: BoundMethodType, args: tuple, kwargs: dict, save_state: bool):
        super(CachingMethodWrapper, self).__init__(method=method, args=args, kwargs=kwargs)
        self._save_state: bool = save_state

        state_string = hash_any(self._get_instance_state()) if self._save_state else "ignore_state"
        self._path_to_cached_method = CACHE_DIR.joinpath(
            # path to module
            *type(self._instance).__module__.split("."),
            # class name
            type(self._instance).__name__,
            # instance state hash
            state_string,
            # instance method name
            method.__name__,
            # args and kwargs hash
            hash_any((args, kwargs)),
        )
        self._path_to_state = self._path_to_cached_method.joinpath("state.dill")
        self._path_to_result = self._path_to_cached_method.joinpath("result.dill")

    @property
    def _instance(self) -> BoundMethodInstanceType:
        return self._method.__self__

    def _get_instance_state(self) -> dict:
        if hasattr(self._instance, "__getstate__"):
            state: dict = getattr(self._instance, "__getstate__")()
        else:
            state = self._instance.__dict__
        return state

    def _set_instance_state(self, state: dict) -> None:
        if hasattr(self._instance, "__setstate__"):
            getattr(self._instance, "__setstate__")(state)
        else:
            self._instance.__dict__.update(state)

    def run_wrapped(self) -> BoundMethodReturnType:
        if not self._path_to_cached_method.exists():
            result: BoundMethodReturnType = self._method(*self._args, **self._kwargs)
            logger.info(f"Saving cache to: {self._path_to_cached_method}")
            self._path_to_cached_method.mkdir(parents=True, exist_ok=False)
            if self._save_state:
                self._path_to_state.write_bytes(dill.dumps(self._get_instance_state()))
            self._path_to_result.write_bytes(dill.dumps(result))
            return result
        else:
            logger.info(f"Loading cache from: {self._path_to_cached_method}")
            if self._save_state:
                self._set_instance_state(dill.loads(self._path_to_state.read_bytes()))
            result = dill.loads(self._path_to_result.read_bytes())
        return result


def _caching_method_wrapper(
    method: BoundMethodType, instance: BoundMethodInstanceType, args: tuple, kwargs: dict, save_state: bool
) -> BoundMethodReturnType:
    return CachingMethodWrapper(method=method, args=args, kwargs=kwargs, save_state=save_state).run_wrapped()


def patch_method_by_name(name: str, cls: type) -> None:
    pass


def patch_method_with_caching(name: str, cls: BoundMethodClassType, save_state: bool) -> None:
    wrapt.wrap_function_wrapper(
        module=cls.__module__,
        name=f"{cls.__name__}.{name}",
        wrapper=partial(_caching_method_wrapper, save_state=save_state),
    )
